#pragma once

// SC - server to client
// CS - client to server

#include<string.h>
#include<stdint.h>

#define PKT_PERFRAME (1<<15)

#define CMD_SC_ASSIGN 0
#define CMD_SC_SINGLE_COMPONENT 1
#define CMD_SC_POSUPDATE 2
#define CMD_SC_PING 3
#define CMD_SC_LOAD_SCRIPT 4
#define CMD_SC_MSG 5

#define CMD_CTYPE_PHYSICS 0
#define CMD_CTYPE_RENDER 1
#define CMD_CTYPE_MOVEMENT 2
#define CMD_CTYPE_PLAYERCTRL 3
#define CMD_CTYPE_BONED 3

#define PEER_DISCONNECT_TOO_MANY_PLAYERS 1

struct bstr {
	uint32_t cap;
	uint32_t wi;
	uint32_t ri;
	uint8_t *data;
};

static inline void bstr(struct bstr *dis, size_t cap) {
	dis->cap = cap;
	dis->wi = dis->ri = 0;
	dis->data = malloc(dis->cap);
}

static inline void bwrap(struct bstr *dis, uint8_t *data, size_t dataLength) {
	dis->cap = dataLength;
	dis->wi = dis->ri = 0;
	dis->data = data;
}

static inline void b_inc(struct bstr *b, uint32_t i) {
	if(b->wi + i > b->cap) {
		b->data = realloc(b->data, b->cap = ((b->cap + i + 63) & ~63));
	}
}

static inline void b_wf32(struct bstr *b, float f) {
	b_inc(b, sizeof(f));
	memcpy(b->data + b->wi, &f, sizeof(f));
	b->wi += sizeof(f);
}

static inline void b_wf64(struct bstr *b, double f) {
	b_inc(b, sizeof(f));
	memcpy(b->data + b->wi, &f, sizeof(f));
	b->wi += sizeof(f);
}

static inline void b_wu8(struct bstr *b, uint8_t i) {
	b_inc(b, sizeof(i));
	b->data[b->wi] = i;
	b->wi += sizeof(i);
}

static inline void b_wu16(struct bstr *b, uint16_t i) {
	b_inc(b, sizeof(i));
	memcpy(b->data + b->wi, &i, sizeof(i));
	b->wi += sizeof(i);
}

static inline void b_wu32(struct bstr *b, uint32_t i) {
	b_inc(b, sizeof(i));
	memcpy(b->data + b->wi, &i, sizeof(i));
	b->wi += sizeof(i);
}

static inline void b_wu64(struct bstr *b, uint64_t i) {
	b_inc(b, sizeof(i));
	memcpy(b->data + b->wi, &i, sizeof(i));
	b->wi += sizeof(i);
}

static inline void b_wf16(struct bstr *b, float f) {
	uint32_t I = *(uint32_t*) &f;
	
	uint16_t i = (I >> 16) & 0x8000;
	int32_t mant = I & 0x7FFFFF;
	uint16_t exp = (I >> 23) & 0xFF;
	
	if(exp == 0xFF) {
		i |= mant ? 0x7E00 : 0x7C00;
	} else if(exp >= 0x8F) {
		i |= 0x7C00;
	} else if(exp >= 0x71) {
		i |= ((exp - 0x70) << 10) | (mant >> 13);
	} else if(exp >= 0x67) {
		i |= (mant | 0x800000) >> (0x7E - exp);
	}
	
	b_wu16(b, i);
}

// TODO: implement varints
static inline void b_wvu(struct bstr *b, size_t i) {
}

static inline void b_wzstr(struct bstr *b, const char *str) {
	size_t l = strlen(str);
	b_inc(b, l + 1);
	strcpy((char*) b->data + b->wi, str);
	b->wi += l;
	b->data[b->wi++] = 0;
}

static inline void b_wraw(struct bstr *b, const uint8_t *data, size_t len) {
	b_inc(b, len);
	memcpy(b->data + b->wi, data, len);
	b->wi += len;
}

static inline float b_rf32(struct bstr *b) {
	float ret;
	memcpy(&ret, b->data + b->ri, sizeof(ret));
	b->ri += sizeof(ret);
	return ret;
}

static inline float b_rf64(struct bstr *b) {
	double ret;
	memcpy(&ret, b->data + b->ri, sizeof(ret));
	b->ri += sizeof(ret);
	return ret;
}

static inline uint16_t b_ru8(struct bstr *b) {
	uint8_t ret;
	memcpy(&ret, b->data + b->ri, sizeof(ret));
	b->ri += sizeof(ret);
	return ret;
}

static inline uint16_t b_ru16(struct bstr *b) {
	uint16_t ret;
	memcpy(&ret, b->data + b->ri, sizeof(ret));
	b->ri += sizeof(ret);
	return ret;
}

static inline uint16_t b_ru32(struct bstr *b) {
	uint32_t ret;
	memcpy(&ret, b->data + b->ri, sizeof(ret));
	b->ri += sizeof(ret);
	return ret;
}

static inline uint16_t b_ru64(struct bstr *b) {
	uint64_t ret;
	memcpy(&ret, b->data + b->ri, sizeof(ret));
	b->ri += sizeof(ret);
	return ret;
}

static inline float b_rf16(struct bstr *b) {
	uint16_t i = b_ru16(b);
	
	uint32_t I = (i & 0x8000) << 16;
	uint32_t mant = i & 0x3FF;
	uint32_t exp = (i & 0x7C00) >> 10;
	if(exp == 0x1F) {
		I |= mant ? 0x7FC00000 : 0x7F800000;
	} else if(exp > 0) {
		I |= (exp + 0x70) << 23;
		if(mant) {
			I |= mant << 13;
		}
	} else if(mant) {
		for(int j = 9; j >= 0; j--) {
			if(mant & (1 << j)) {
				I |= ((0x67 + j) << 23) | ((mant << (23 - j)) & 0x7FFFFF);
				break;
			}
		}
	}
	
	return *(float*) &I;
}

static inline const char *b_rzstr(struct bstr *b) {
	const char *ret = (char*) b->data + b->ri;
	b->ri += strlen(ret) + 1;
	return ret;
}

static inline const uint8_t *b_rraw(struct bstr *b, size_t len) {
	const uint8_t *ret = b->data + b->ri;
	b->ri += len;
	return ret;
}

static inline void b_rraw_copy(struct bstr *b, uint8_t *data, size_t len) {
	memcpy(data, b_rraw(b, len), len);
}
