#include"resman.h"

#include<string.h>
#include<stdlib.h>
#include<stdio.h>
#include"k3.h"

struct ResMan ResMan;

static uint64_t fnv1a(const char *str) {
	uint64_t ret = 14695981039346656037ULL;
	
	while(*str) {
		ret ^= *(const uint8_t*) str;
		ret *= 1099511628211ULL;
		
		str++;
	}
	
	return ret;
}

static struct ResManRes *make(enum ResManType type, const char *name, size_t refs) {
	struct ResManRes *ret = malloc(sizeof(*ret));
	ret->type = type;
	ret->name = strdup(name);
	ret->refs = refs;
	ret->next = NULL;
	ret->thing = NULL;
	
	if(type != RESMAN_OTHER) {
		k3Log(k3_INFO, "Loading %s...", name);
		if(!ResMan.handlers[type].loader(ResMan.handlers[type].ud, ret)) {
			printf("Failed to load resource %s (%i)\n", name, type);
		}
	}
	
	return ret;
}

struct ResManRes *resman_reffull(enum ResManType type, const char *name) {
	size_t bucket = fnv1a(name) % RESMAN_BUCKETS;
	
	struct ResManRes *ptr = ResMan.buckets[bucket];
	
	if(!ptr) {
		return (ResMan.buckets[bucket] = make(type, name, 1));
	}
	
	struct ResManRes *prev = NULL;
	while(ptr) {
		if(ptr->type == type && !strcmp(name, ptr->name)) {
			ptr->refs++;
			return ptr;
		}
		
		prev = ptr;
		ptr = ptr->next;
	}
	
	return (prev->next = make(type, name, 1));
}

void *resman_ref(enum ResManType type, const char *name) {
	return resman_reffull(type, name)->thing;
}

void resman_unref(enum ResManType type, const char *name) {
	size_t bucket = fnv1a(name) % RESMAN_BUCKETS;
	
	struct ResManRes *prev = NULL;
	struct ResManRes *ptr = ResMan.buckets[bucket];
	
	if(!ptr) goto notfound;
	
	while(ptr) {
		if(ptr->type == type && !strcmp(name, ptr->name)) {
			ptr->refs--;
			if(ptr->refs == 0) {
				if(ResMan.handlers[type].unloader) {
					ResMan.handlers[type].unloader(ResMan.handlers[type].ud, name, ptr->thing);
				} else {
					printf("Memory leak! No unloader for (%i)!\n", type);
				}
				
				if(prev) {
					prev->next = ptr->next;
				} else {
					ResMan.buckets[bucket] = ptr->next;
				}
				
				free(ptr->name);
				free(ptr);
			}
			
			return;
		}
		
		prev = ptr;
		ptr = ptr->next;
	}
	
notfound:
	printf("Attempt to release unallocated resource %s (%i)!\n", name, type);
}

void resman_unref_ptr(enum ResManType type, void *v) {
	for(size_t bucket = 0; bucket < RESMAN_BUCKETS; bucket++) {
		struct ResManRes *prev = NULL;
		struct ResManRes *ptr = ResMan.buckets[bucket];
		
		while(ptr) {
			if(ptr->thing == v) {
				ptr->refs--;
				if(ptr->refs == 0) {
					if(ResMan.handlers[type].unloader) {
						ResMan.handlers[type].unloader(ResMan.handlers[type].ud, ptr->name, ptr->thing);
					} else {
						printf("Memory leak! No unloader for (%i)!\n", type);
					}
					
					if(prev) {
						prev->next = ptr->next;
					} else {
						ResMan.buckets[bucket] = ptr->next;
					}
					
					free(ptr->name);
					free(ptr);
				}
				
				return;
			}
			
			prev = ptr;
			ptr = ptr->next;
		}
	}
	
notfound:
	printf("Attempt to release unnamed unallocated resource (%i)!\n", type);
}

struct ResManRes *resman_rev(void *v) {
	for(size_t bucket = 0; bucket < RESMAN_BUCKETS; bucket++) {
		struct ResManRes *prev = NULL;
		struct ResManRes *ptr = ResMan.buckets[bucket];
		
		while(ptr) {
			if(ptr->thing == v) {
				return ptr;
			}
			
			prev = ptr;
			ptr = ptr->next;
		}
	}
	
	return NULL;
}
